"""
Includes Input/Output (I/O) functionalities like reading and writing from and into specific file formats.
"""

import json
import onnx

import pandas as pd


def read_json(input_path):
    """
    Reads the ``json`` file of the given ``input_path``.

    :param input_path: Path to the json file
    :type input_path: str
    :return: A loaded json object.
    """
    with open(input_path, encoding='utf-8') as f:
        json_data = json.load(f)

    return json_data


def read_df_from_json(input_path, key=None):
    """
    Reads the ``json`` file of the given ``input_path`` and converts it to pandas ``dataframe``.
    :param input_path: Path to the json file
    :type input_path: str
    :param key: Specifies the object to be converted.
    :type key: str
    :return: A loaded pandas dataframe object.
    """
    json_file = read_json(input_path)

    if key:
        return pd.json_normalize(json_file[key])

    return pd.json_normalize(json_file)


def read_onnx(input_path):
    """
    Reads the ``onnx`` file of the given ``input_path``.

    :param input_path: Path to the onnx file
    :type input_path: str
    :return: A loaded onnx object.
    """
    return onnx.load(input_path)



"""
Includes custom-defined exceptions
"""


class ORKGNLPValidationError(Exception):
    """
    Indicates an incorrect input attempt by the caller.
    """

    def __init__(self, message):
        super().__init__(message)



"""
Provides a set of global variables within the orkgnlp_context.
"""

import os

from orkgnlp.util import io
from orkgnlp.util.datastructure import StrictDict

_current_dir = os.path.dirname(os.path.realpath(__file__))

"""
Context dictionary for library's global configuration variables. 
Keys represent environment variables and values represent default values. 
Values can be changed using the corresponding setter function.
"""
orkgnlp_context = StrictDict(
    {
        'ORKG_NLP_DATA_CACHE_ROOT': os.path.join(os.path.expanduser('~'), 'orkgnlp_data'),
        'PREDICATES_CLUSTERING_SERVICE_NAME': 'predicates-clustering',
        'HUGGINGFACE_REPOS': io.read_json(os.path.join(_current_dir, '..', 'huggingface_repos.json'))
    }
)


def set_data_cache_root(cache_root):
    """
    Overrides the value of ORKG_NLP_DATA_CACHE_ROOT. See :doc:`../configure`

    :param cache_root: Path to the data cache root. The path must be absolute.
    :type cache_root: str
    """
    orkgnlp_context['ORKG_NLP_DATA_CACHE_ROOT'] = cache_root



""" Predicates recommendation service. """
import numpy as np
import onnxruntime as rt

from orkgnlp.clustering._predicates_config import config
from orkgnlp.tools import downloader
from orkgnlp.util import io, text
from orkgnlp.util.decorators import singleton


@singleton
class PredicatesRecommender:
    """
    The PredicatesRecommender follows the singleton pattern, i.e. only one instance can be obtained from it.

    It requires a clustering model, vectorizer, training set and predicates. The required files are downloaded while
    initiation, if it has not happened before.

    :note: This class documentation will not be generated with sphinx :autosummary:, because the @singleton decorator
    returns a function that cannot be detected as a class. TODO: fix this issue!
    """

    def __init__(self):
        downloader.exists_or_download(config['service_name'])

        self._model = io.read_onnx(config['paths']['model'])
        self._vectorizer = io.read_onnx(config['paths']['vectorizer'])
        self._train_df = io.read_df_from_json(config['paths']['training_data'], key='instances')
        self._predicates = io.read_json(config['paths']['predicates'])

    def recommend(self, title, abstract):
        """
        Recommends predicates for a research paper.

        :param title: Title of the research paper.
        :type title: str
        :param abstract: Abstract of the research paper.
        :type abstract: str
        :return: List of predicates.
        """
        return self._recommend(q='{} {}'.format(title, abstract))

    def _recommend(self, q):
        vectorized_text = self._vectorize_input(q)

        comparison_ids = self._predict_comparisons(vectorized_text)
        return self._map_to_predicates(comparison_ids)

    def _vectorize_input(self, q):
        preprocessed_text = self._text_process(q)
        session = rt.InferenceSession(self._vectorizer.SerializeToString())
        output = session.run(['variable'], {session.get_inputs()[0].name: [[preprocessed_text]]})
        return output[0][0]

    def _predict_comparisons(self, vectorized_text):
        session = rt.InferenceSession(self._model.SerializeToString())
        output = session.run(['label', 'labels_'], {session.get_inputs()[0].name: [vectorized_text]})
        cluster_label, model_labels_ = output[0], output[1]
        cluster_instances_indices = np.argwhere(cluster_label == model_labels_).squeeze(1)
        cluster_instances = self._train_df.iloc[cluster_instances_indices]
        comparison_ids = cluster_instances['comparison_id'].unique()
        return comparison_ids

    def _map_to_predicates(self, comparison_ids):
        predicate_ids = []
        predicates = []

        for comparison_id in comparison_ids:
            for predicate in self._predicates[comparison_id]:
                if predicate['id'] in predicate_ids:
                    continue

                predicate_ids.append(predicate['id'])
                predicates.append(predicate)

        return predicates

    @staticmethod
    def _text_process(q):
        q = text.remove_punctuation(q)
        q = text.remove_stopwords(q)
        q = text.lemmatize(q)
        return q.lower()



"""
Root package of orkgnlp.
"""

__version__ = '0.1.6'

import logging

from orkgnlp.tools import *

# Root logger configuration
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

stdout = logging.StreamHandler()
stdout.setLevel(logging.DEBUG)

formatter = logging.Formatter('%(name)s - %(levelname)s: %(message)s')
stdout.setFormatter(formatter)

logger.addHandler(stdout)



""" Config file for the predicates clustering service. """

import os

from orkgnlp.config import orkgnlp_context
from orkgnlp.util.datastructure import StrictDict

_service_data_dir = os.path.join(
    orkgnlp_context.get('ORKG_NLP_DATA_CACHE_ROOT'),
    orkgnlp_context.get('PREDICATES_CLUSTERING_SERVICE_NAME')
)
_service_data_files = \
    orkgnlp_context.get('HUGGINGFACE_REPOS')[orkgnlp_context.get('PREDICATES_CLUSTERING_SERVICE_NAME')][0]['files']

config = StrictDict({
    'service_name': orkgnlp_context.get('PREDICATES_CLUSTERING_SERVICE_NAME'),
    'paths': {
        'vectorizer': os.path.join(_service_data_dir, _service_data_files['vectorizer']),
        'model': os.path.join(_service_data_dir, _service_data_files['model']),
        'training_data': os.path.join(_service_data_dir, _service_data_files['training_data']),
        'predicates': os.path.join(_service_data_dir, _service_data_files['predicates'])
    }
})



class BaseAnnotator:

    @staticmethod
    def annotate(**kwargs):
        pass


class CSTitleAnnotator(BaseAnnotator):

    @staticmethod
    def annotate(text):
        pass


class CSAbstractAnnotator(BaseAnnotator):

    @staticmethod
    def annotate(text):
        pass


class CSAnnotator(BaseAnnotator):

    @staticmethod
    def annotate(title, abstract):
        return {
            'title_entities': CSAbstractAnnotator.annotate(text=title),
            'abstract_entities': CSAnnotator.annotate(text=abstract)
        }



"""
Includes help classes useful for encapsulating data.
"""


class StrictDict(dict):
    """
    This class prevents adding new keys to the dictionary
    """

    def __setitem__(self, key, value):
        self.check(key)
        dict.__setitem__(self, key, value)

    def update(self, __m, **kwargs):
        for key, value in __m.items():
            self.check(key)

        for key, value in __m.items():
            self.__setitem__(key, value)

    def check(self, key):
        if key not in self:
            raise KeyError("{} is not a legal key of this StrictDict".format(repr(key)))



""" Utility file for text processing functionalities. """
import string

import nltk
from nltk.stem import WordNetLemmatizer
from nltk.corpus import stopwords


def remove_punctuation(text):
    """
    Removes punctuations from the given ``text``.

    :param text:
    :type text: str
    :return: text with punctuations removed.
    """
    return text.translate(str.maketrans('', '', string.punctuation))


def remove_stopwords(text, language='english'):
    """
    Removes stopwords from the given ``text``.
    Uses the stopwords defined in ``NLTK``.

    :param text:
    :type text: str
    :param language: defaults to ``english``.
    :type language: str
    :return: text with stopwords removed.
    """
    nltk.download('stopwords', quiet=True)

    return ' '.join(
        [word for word in text.split() if word not in stopwords.words(language)]
    )


def lemmatize(text):
    """
    Lemmatizes the words in the given ``text``.
    Uses the WordNetLemmatizer defined in ``NLTK``.

    :param text:
    :type text: str
    :return: text with its lemmatized words.
    """
    nltk.download('wordnet', quiet=True)

    stemmer = WordNetLemmatizer()
    return ' '.join(
        [stemmer.lemmatize(word) for word in text.split()]
    )



"""
A script for downloading ORKG-NLP models and data needed to use the supported :doc:`../services/services`.

This script depends mainly on the ``huggingface_hub`` client and fetches the files from our
huggingface `repositories <https://huggingface.co/orkg>`_.
"""

import os
import logging
from huggingface_hub import hf_hub_download

from orkgnlp.config import orkgnlp_context
from orkgnlp.util.exceptions import ORKGNLPValidationError

logger = logging.getLogger(__name__)


def _services_are_known(services, orkg_services):
    """
    Returns True if the given services names are subset of the known ORKG-NLP services

    :param services: list of service names
    :type services: list[str]
    :param orkg_services: list of ORKG-NLP service names
    :type services: list[str]
    :return:
    """
    return set(services).issubset(orkg_services)


def download(services):
    """
    Downloads the models and data needed to use the supported services.

    The download destination is given by ``ORKG_NLP_DATA_CACHE_ROOT``.
    You can also check how to :doc:`../configure` its value.

    :param services: a string representing a service name or a list of them. Check :doc:`../services/services` for a full list.
    :type services: str or list[str]
    :raise orkgnlp.util.exceptions.ORKGNLPValidationError: If one of the known passed service names is unknown.
    """
    if isinstance(services, str):
        services = [services]

    orkg_services = orkgnlp_context.get('HUGGINGFACE_REPOS')
    if not _services_are_known(services, orkg_services.keys()):
        raise ORKGNLPValidationError('Unknown model name(s) given {}. Please check the following known services: {}'
                              .format(services, list(orkg_services.keys())))

    cache_root = orkgnlp_context.get('ORKG_NLP_DATA_CACHE_ROOT')
    logger.info('Downloading to {}'.format(cache_root))
    for service in services:
        for repo in orkg_services[service]:
            for filename in repo['files'].values():
                hf_hub_download(
                    repo_id=repo['repo_id'],
                    filename=filename,
                    force_filename=filename,
                    cache_dir=os.path.join(cache_root, service)
                )


def exists_or_download(service):
    """
    Checks the presence of the required files for executing the given service and downloads them in case of absence.

    :param service: a string representing a ORKG-NLP service name. Check :doc:`../services/services` for a full list.
    :type service: str
    :return:
    """
    service_dir = os.path.join(orkgnlp_context.get('ORKG_NLP_DATA_CACHE_ROOT'), service)
    if not os.path.exists(service_dir):
        download(service)



""" Defines decorators used within the package """


def singleton(cls):
    """
    Reflects the singleton pattern and can be used to decorate classes.

    Only one instance can be obtained from a singleton class.

    :param cls: The decorated class.
    :return: Function responsible for returning a pre-initiated instance of the given ``cls``.
    """

    instances = {}

    def getinstance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]

    return getinstance



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['orkgnlp',
 'orkgnlp.annotation',
 'orkgnlp.annotation.csner',
 'orkgnlp.annotation.pdf',
 'orkgnlp.annotation.sentence',
 'orkgnlp.annotation.summary',
 'orkgnlp.annotation.table',
 'orkgnlp.annotation.text',
 'orkgnlp.clustering',
 'orkgnlp.config',
 'orkgnlp.tools',
 'orkgnlp.util']

package_data = \
{'': ['*']}

install_requires = \
['huggingface-hub>=0.5.1,<0.6.0',
 'nltk==3.5',
 'onnx==1.11.0',
 'onnxruntime==1.11.1',
 'pandas==1.4.2']

setup_kwargs = {
    'name': 'orkgnlp',
    'version': '0.1.7',
    'description': 'Python package wrapping the ORKG NLP Services.',
    'long_description': '# ORKG NLP PyPI\n[![Documentation Status](https://readthedocs.org/projects/orkg-nlp-pypi/badge/?version=latest)](https://orkg-nlp-pypi.readthedocs.io/en/latest/?badge=latest)\n[![PyPI version](https://badge.fury.io/py/orkgnlp.svg)](https://badge.fury.io/py/orkgnlp)\n\nPyPI package wrapping the ORKG NLP services.\n\nCheck our [Read the Docs](https://orkg-nlp-pypi.readthedocs.io/en/latest/) for more details!',
    'author': 'Omar Arab Oghli',
    'author_email': 'omar.araboghli@outlook.com',
    'maintainer': 'Omar Arab Oghli',
    'maintainer_email': None,
    'url': 'http://orkg.org/about',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
